/*
 * Copyright (c) 2007 - 2015 Joseph Gaeddert
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//
// firdecim_crcf_data_M2h4x20.c: autotest firdecim data
//

#include <complex.h>

float firdecim_crcf_data_M2h4x20_h[] = {
    0.031235809409,
   -0.216021656042,
   -0.123125426658,
    0.014724190886};

float complex firdecim_crcf_data_M2h4x20_x[] = {
   -0.121360059018 +  -0.040946328225*_Complex_I,
   -0.048183917318 +  -0.090893384286*_Complex_I,
    0.050431539967 +  -0.039477767743*_Complex_I,
   -0.092873851447 +  -0.016290157417*_Complex_I,
    0.067927836995 +   0.034766499357*_Complex_I,
   -0.062738323035 +  -0.070124552362*_Complex_I,
   -0.089685845706 +   0.085063483262*_Complex_I,
    0.036320846361 +   0.192249836854*_Complex_I,
   -0.097540647206 +  -0.078675689987*_Complex_I,
    0.062450869382 +   0.091837474249*_Complex_I,
   -0.062450045345 +  -0.029363766863*_Complex_I,
   -0.030191356564 +   0.094898259922*_Complex_I,
    0.130002759179 +  -0.034950407712*_Complex_I,
    0.030652983878 +   0.015398314893*_Complex_I,
   -0.028490871597 +   0.130293581637*_Complex_I,
    0.114396819866 +  -0.058213046230*_Complex_I,
   -0.090569184709 +  -0.080687388201*_Complex_I,
   -0.002440656567 +   0.093805498501*_Complex_I,
    0.181551012573 +   0.015635958339*_Complex_I,
    0.081568978667 +   0.051967040191*_Complex_I};

float complex firdecim_crcf_data_M2h4x20_y[] = {
   -0.003790779673 +  -0.001278991704*_Complex_I,
    0.026926548630 +   0.023443353500*_Complex_I,
    0.015265670091 +   0.008127371986*_Complex_I,
    0.001020290227 +   0.013284949227*_Complex_I,
   -0.000774013472 +  -0.055493631953*_Complex_I,
   -0.002896889061 +  -0.008238383106*_Complex_I,
    0.019191455251 +  -0.016624124713*_Complex_I,
   -0.023962832263 +   0.006444039955*_Complex_I,
   -0.023581901158 +  -0.005760782332*_Complex_I,
    0.019033897622 +  -0.010698088224*_Complex_I};

