/*
 * Copyright (c) 2007 - 2015 Joseph Gaeddert
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//
// firfilt_crcf_data_h4x8.c: autotest firfilt data
//

#include <complex.h>

float firfilt_crcf_data_h4x8_h[] = {
    0.081125556518,
   -0.048097526791,
    0.083945750272,
    0.016820374297};

float complex firfilt_crcf_data_h4x8_x[] = {
    0.212556497097 +  -0.062593316778*_Complex_I,
    0.074632428892 +  -0.155168218555*_Complex_I,
   -0.139928836211 +  -0.151937330426*_Complex_I,
    0.057785165005 +   0.027214057011*_Complex_I,
    0.047104310251 +  -0.016114794312*_Complex_I,
    0.140236563713 +   0.296500956218*_Complex_I,
   -0.134588116658 +  -0.011633439603*_Complex_I,
    0.035369414992 +  -0.047878728319*_Complex_I};

float complex firfilt_crcf_data_h4x8_y[] = {
    0.017243764119 +  -0.005077917658*_Complex_I,
   -0.004168844485 +  -0.009577524353*_Complex_I,
    0.002901774665 +  -0.010117235877*_Complex_I,
    0.021258439696 +  -0.004562990201*_Complex_I,
   -0.009449045890 +  -0.017980731205*_Complex_I,
    0.011608332082 +   0.024557748499*_Complex_I,
   -0.012737392976 +  -0.016099749821*_Complex_I,
    0.021907294708 +   0.021294289547*_Complex_I};

