/*
 * Copyright (c) 2007 - 2015 Joseph Gaeddert
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//
// firdecim_rrrf_data_M3h7x30.c: autotest firdecim data
//

float firdecim_rrrf_data_M3h7x30_h[] = {
   -0.074705344704,
    0.020538592337,
   -0.020061495576,
    0.096703320953,
    0.100160540368,
    0.097166994081,
    0.083888510566};

float firdecim_rrrf_data_M3h7x30_x[] = {
   -0.077089593357,
    0.032375219181,
    0.141488397385,
    0.034405467968,
   -0.138389144841,
    0.154646756715,
    0.086331166522,
    0.041332691838,
    0.091791117954,
   -0.065105465675,
    0.075745848821,
    0.012475523017,
   -0.107768670583,
    0.233864631918,
   -0.110944430251,
   -0.060308141718,
   -0.251948448147,
    0.153274083486,
   -0.114102376762,
   -0.183548035449,
   -0.152968693185,
    0.049660839590,
   -0.056844064619,
   -0.153304122180,
    0.018994945270,
    0.000201893929,
   -0.055497296237,
    0.101173781919,
   -0.050453255942,
   -0.019320758294};

float firdecim_rrrf_data_M3h7x30_y[] = {
    0.005759004645,
   -0.007768614835,
    0.013680669269,
    0.019197170461,
    0.020943849985,
   -0.009738613503,
    0.013465672357,
   -0.028391740957,
   -0.041353046024,
   -0.023577651344};

