/*
 * Copyright (c) 2007 - 2015 Joseph Gaeddert
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//
// firfilt_rrrf_data_h7x16.c: autotest firfilt data
//

float firfilt_rrrf_data_h7x16_h[] = {
    0.167653994833,
    0.011522192990,
    0.015882599202,
   -0.093561396312,
    0.074301353348,
   -0.124295601121,
    0.032095823166};

float firfilt_rrrf_data_h7x16_x[] = {
    0.096678620741,
    0.015023630036,
    0.051530016668,
   -0.094237043292,
    0.007423744308,
    0.024664429619,
   -0.199548815687,
    0.003258035222,
    0.082829872734,
    0.142277057992,
   -0.149671886106,
   -0.058687993370,
    0.018936199920,
   -0.026660942281,
    0.016612364911,
    0.178689105006};

float firfilt_rrrf_data_h7x16_y[] = {
    0.016208556982,
    0.003632721318,
    0.010347826097,
   -0.024012250428,
    0.006754954149,
   -0.012997772766,
   -0.019171751321,
   -0.014980556455,
    0.022366048926,
    0.041414757082,
   -0.040097174540,
    0.008782967641,
   -0.013845574916,
    0.009200324306,
   -0.017877119698,
    0.046763669812};

