/*
 * Copyright (c) 2007 - 2015 Joseph Gaeddert
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//
// firdecim_rrrf_data_M2h4x20.c: autotest firdecim data
//

float firdecim_rrrf_data_M2h4x20_h[] = {
    0.070782309250,
   -0.114245586699,
    0.008215220382,
   -0.116718567037};

float firdecim_rrrf_data_M2h4x20_x[] = {
    0.098311328839,
   -0.051405684088,
   -0.097846553674,
    0.070744555579,
   -0.092569362020,
    0.125743342560,
   -0.015936561579,
    0.092440100110,
   -0.007046555447,
   -0.011409188742,
    0.065796277764,
    0.121624056526,
    0.001291608019,
   -0.103251042400,
   -0.067174551288,
   -0.069262060525,
    0.038107684731,
   -0.091432498375,
    0.094339814736,
    0.004969612518};

float firdecim_rrrf_data_M2h4x20_y[] = {
    0.006958702881,
   -0.000245283250,
   -0.009438359685,
   -0.024511329437,
   -0.025867150062,
   -0.004886703086,
   -0.011931393613,
   -0.007143968697,
    0.022109694637,
    0.025520580844};

