/*
 * Copyright (c) 2007 - 2015 Joseph Gaeddert
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//
// firfilt_rrrf_data_h4x8.c: autotest firfilt data
//

float firfilt_rrrf_data_h4x8_h[] = {
   -0.121887159208,
   -0.231303112477,
   -0.011081038093,
    0.002940945390};

float firfilt_rrrf_data_h4x8_x[] = {
    0.087523073427,
    0.091626543082,
    0.069905988906,
   -0.025530671018,
   -0.085926435885,
    0.016121796124,
    0.067241716218,
    0.036151454364};

float firfilt_rrrf_data_h4x8_y[] = {
   -0.010667938785,
   -0.031412458342,
   -0.030683993510,
   -0.013815528486,
    0.015873490575,
    0.018398508167,
   -0.011047853592,
   -0.020390967516};

