/*
 * Copyright (c) 2007 - 2015 Joseph Gaeddert
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

//
// firfilt_cccf_data_h4x8.c: autotest firfilt data
//

#include <complex.h>

float complex firfilt_cccf_data_h4x8_h[] = {
   -0.031084031859 +  -0.013523088141*_Complex_I,
   -0.129175459526 +   0.067705781544*_Complex_I,
    0.221460930668 +  -0.057515301525*_Complex_I,
   -0.023734659272 +   0.094741117144*_Complex_I};

float complex firfilt_cccf_data_h4x8_x[] = {
    0.035529355081 +  -0.025645176804*_Complex_I,
    0.010053327033 +   0.058182925297*_Complex_I,
   -0.051911580279 +  -0.038455783791*_Complex_I,
    0.087876912441 +  -0.008273223193*_Complex_I,
   -0.195891631367 +   0.043381908184*_Complex_I,
   -0.018068686759 +   0.096890714727*_Complex_I,
    0.072440149482 +   0.119581756000*_Complex_I,
   -0.070807340362 +   0.027693599784*_Complex_I};

float complex firfilt_cccf_data_h4x8_y[] = {
   -0.001451197592 +   0.000316688892*_Complex_I,
   -0.002378879142 +   0.003773758321*_Complex_I,
    0.002248990658 +  -0.012660658356*_Complex_I,
    0.013625136077 +   0.016803422304*_Complex_I,
   -0.023574746940 +   0.002359815677*_Complex_I,
    0.048100008285 +  -0.032526172551*_Complex_I,
   -0.047049781567 +   0.010960128197*_Complex_I,
   -0.012767824065 +  -0.007537619160*_Complex_I};

